// <copyright file="ShippingAPIClient.cs" company="APIMatic">
// Copyright (c) APIMatic. All rights reserved.
// </copyright>
namespace ShippingAPI.Standard
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Text;
    using ShippingAPI.Standard.Authentication;
    using ShippingAPI.Standard.Controllers;
    using ShippingAPI.Standard.Http.Client;
    using ShippingAPI.Standard.Utilities;

    /// <summary>
    /// The gateway for the SDK. This class acts as a factory for Controller and
    /// holds the configuration of the SDK.
    /// </summary>
    public sealed class ShippingAPIClient : IConfiguration
    {
        // A map of environments and their corresponding servers/baseurls
        private static readonly Dictionary<Environment, Dictionary<Server, string>> EnvironmentsMap =
            new Dictionary<Environment, Dictionary<Server, string>>
        {
            {
                Environment.Production, new Dictionary<Server, string>
                {
                    { Server.YourInstance, "https://{InstanceURL}/Logistics%20Server/" },
                }
            },
        };

        private readonly IDictionary<string, IAuthManager> authManagers;
        private readonly IHttpClient httpClient;
        private readonly CustomQueryAuthenticationManager customQueryAuthenticationManager;

        private readonly Lazy<CommonController> common;

        private ShippingAPIClient(
            string contentType,
            Environment environment,
            string instanceURL,
            string apikey,
            IDictionary<string, IAuthManager> authManagers,
            IHttpClient httpClient,
            IHttpClientConfiguration httpClientConfiguration)
        {
            this.ContentType = contentType;
            this.Environment = environment;
            this.InstanceURL = instanceURL;
            this.httpClient = httpClient;
            this.authManagers = (authManagers == null) ? new Dictionary<string, IAuthManager>() : new Dictionary<string, IAuthManager>(authManagers);
            this.HttpClientConfiguration = httpClientConfiguration;

            this.common = new Lazy<CommonController>(
                () => new CommonController(this, this.httpClient, this.authManagers));

            if (this.authManagers.ContainsKey("global"))
            {
                this.customQueryAuthenticationManager = (CustomQueryAuthenticationManager)this.authManagers["global"];
            }

            if (!this.authManagers.ContainsKey("global")
                || !this.CustomQueryAuthenticationCredentials.Equals(apikey))
            {
                this.customQueryAuthenticationManager = new CustomQueryAuthenticationManager(apikey);
                this.authManagers["global"] = this.customQueryAuthenticationManager;
            }
        }

        /// <summary>
        /// Gets CommonController controller.
        /// </summary>
        public CommonController CommonController => this.common.Value;

        /// <summary>
        /// Gets the configuration of the Http Client associated with this client.
        /// </summary>
        public IHttpClientConfiguration HttpClientConfiguration { get; }

        /// <summary>
         /// Gets ContentType.
        /// </summary>
        public string ContentType { get; }

        /// <summary>
        /// Gets Environment.
        /// Current API environment.
        /// </summary>
        public Environment Environment { get; }

        /// <summary>
        /// Gets InstanceURL.
        /// Replace yourinstance.transtream.com with the URL of your instance. You will also need to change the API Key for the call be successful.
        /// </summary>
        public string InstanceURL { get; }

        /// <summary>
        /// Gets auth managers.
        /// </summary>
        internal IDictionary<string, IAuthManager> AuthManagers => this.authManagers;

        /// <summary>
        /// Gets http client.
        /// </summary>
        internal IHttpClient HttpClient => this.httpClient;

        /// <summary>
        /// Gets the credentials to use with CustomQueryAuthentication.
        /// </summary>
        public ICustomQueryAuthenticationCredentials CustomQueryAuthenticationCredentials => this.customQueryAuthenticationManager;

        /// <summary>
        /// Gets the URL for a particular alias in the current environment and appends
        /// it with template parameters.
        /// </summary>
        /// <param name="alias">Default value:YOUR INSTANCE.</param>
        /// <returns>Returns the baseurl.</returns>
        public string GetBaseUri(Server alias = Server.YourInstance)
        {
            StringBuilder url = new StringBuilder(EnvironmentsMap[this.Environment][alias]);
            ApiHelper.AppendUrlWithTemplateParameters(url, this.GetBaseUriParameters());

            return url.ToString();
        }

        /// <summary>
        /// Creates an object of the ShippingAPIClient using the values provided for the builder.
        /// </summary>
        /// <returns>Builder.</returns>
        public Builder ToBuilder()
        {
            Builder builder = new Builder()
                .ContentType(this.ContentType)
                .Environment(this.Environment)
                .InstanceURL(this.InstanceURL)
                .CustomQueryAuthenticationCredentials(this.customQueryAuthenticationManager.Apikey)
                .HttpClient(this.httpClient)
                .AuthManagers(this.authManagers)
                .HttpClientConfig(config => config.Build());

            return builder;
        }

        /// <inheritdoc/>
        public override string ToString()
        {
            return
                $"ContentType = {this.ContentType}, " +
                $"Environment = {this.Environment}, " +
                $"InstanceURL = {this.InstanceURL}, " +
                $"HttpClientConfiguration = {this.HttpClientConfiguration}, ";
        }

        /// <summary>
        /// Creates the client using builder.
        /// </summary>
        /// <returns> ShippingAPIClient.</returns>
        internal static ShippingAPIClient CreateFromEnvironment()
        {
            var builder = new Builder();

            string contentType = System.Environment.GetEnvironmentVariable("SHIPPING_API_STANDARD_CONTENT_TYPE");
            string environment = System.Environment.GetEnvironmentVariable("SHIPPING_API_STANDARD_ENVIRONMENT");
            string instanceURL = System.Environment.GetEnvironmentVariable("SHIPPING_API_STANDARD_INSTANCE_URL");
            string apikey = System.Environment.GetEnvironmentVariable("SHIPPING_API_STANDARD_APIKEY");

            if (contentType != null)
            {
                builder.ContentType(contentType);
            }

            if (environment != null)
            {
                builder.Environment(ApiHelper.JsonDeserialize<Environment>($"\"{environment}\""));
            }

            if (instanceURL != null)
            {
                builder.InstanceURL(instanceURL);
            }

            if (apikey != null)
            {
                builder.CustomQueryAuthenticationCredentials(apikey);
            }

            return builder.Build();
        }

        /// <summary>
        /// Makes a list of the BaseURL parameters.
        /// </summary>
        /// <returns>Returns the parameters list.</returns>
        private List<KeyValuePair<string, object>> GetBaseUriParameters()
        {
            List<KeyValuePair<string, object>> kvpList = new List<KeyValuePair<string, object>>()
            {
                new KeyValuePair<string, object>("InstanceURL", this.InstanceURL),
            };
            return kvpList;
        }

        /// <summary>
        /// Builder class.
        /// </summary>
        public class Builder
        {
            private string contentType = "application/x-www-form-urlencoded";
            private Environment environment = ShippingAPI.Standard.Environment.Production;
            private string instanceURL = "yourinstance.transtream.com";
            private string apikey = "<Obtain API Key>";
            private IDictionary<string, IAuthManager> authManagers = new Dictionary<string, IAuthManager>();
            private HttpClientConfiguration.Builder httpClientConfig = new HttpClientConfiguration.Builder();
            private IHttpClient httpClient;

            /// <summary>
            /// Sets credentials for CustomQueryAuthentication.
            /// </summary>
            /// <param name="apikey">Apikey.</param>
            /// <returns>Builder.</returns>
            public Builder CustomQueryAuthenticationCredentials(string apikey)
            {
                this.apikey = apikey ?? throw new ArgumentNullException(nameof(apikey));
                return this;
            }

            /// <summary>
            /// Sets ContentType.
            /// </summary>
            /// <param name="contentType"> ContentType. </param>
            /// <returns> Builder. </returns>
            public Builder ContentType(string contentType)
            {
                this.contentType = contentType ?? throw new ArgumentNullException(nameof(contentType));
                return this;
            }

            /// <summary>
            /// Sets Environment.
            /// </summary>
            /// <param name="environment"> Environment. </param>
            /// <returns> Builder. </returns>
            public Builder Environment(Environment environment)
            {
                this.environment = environment;
                return this;
            }

            /// <summary>
            /// Sets InstanceURL.
            /// </summary>
            /// <param name="instanceURL"> InstanceURL. </param>
            /// <returns> Builder. </returns>
            public Builder InstanceURL(string instanceURL)
            {
                this.instanceURL = instanceURL ?? throw new ArgumentNullException(nameof(instanceURL));
                return this;
            }

            /// <summary>
            /// Sets HttpClientConfig.
            /// </summary>
            /// <param name="action"> Action. </param>
            /// <returns>Builder.</returns>
            public Builder HttpClientConfig(Action<HttpClientConfiguration.Builder> action)
            {
                if (action is null)
                {
                    throw new ArgumentNullException(nameof(action));
                }

                action(this.httpClientConfig);
                return this;
            }

            /// <summary>
            /// Sets the IHttpClient for the Builder.
            /// </summary>
            /// <param name="httpClient"> http client. </param>
            /// <returns>Builder.</returns>
            internal Builder HttpClient(IHttpClient httpClient)
            {
                this.httpClient = httpClient ?? throw new ArgumentNullException(nameof(httpClient));
                return this;
            }

            /// <summary>
            /// Sets the authentication managers for the Builder.
            /// </summary>
            /// <param name="authManagers"> auth managers. </param>
            /// <returns>Builder.</returns>
            internal Builder AuthManagers(IDictionary<string, IAuthManager> authManagers)
            {
                this.authManagers = authManagers ?? throw new ArgumentNullException(nameof(authManagers));
                return this;
            }

            /// <summary>
            /// Creates an object of the ShippingAPIClient using the values provided for the builder.
            /// </summary>
            /// <returns>ShippingAPIClient.</returns>
            public ShippingAPIClient Build()
            {
                this.httpClient = new HttpClientWrapper(this.httpClientConfig.Build());

                return new ShippingAPIClient(
                    this.contentType,
                    this.environment,
                    this.instanceURL,
                    this.apikey,
                    this.authManagers,
                    this.httpClient,
                    this.httpClientConfig.Build());
            }
        }
    }
}
